import json
import sys

try:
    from importlib.resources import files
except ImportError:  # pragma: no cover
    from importlib_resources import files

from referencing import Registry
from referencing.jsonschema import DRAFT202012, EMPTY_REGISTRY

import referencing_loaders as loaders


def test_absolute_internally_identified(tmp_path):
    root_path, root = tmp_path / "schema.json", {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$id": "http://example.com/",
    }
    son_path, son = tmp_path / "child/son.json", {
        "$schema": "https://json-schema.org/draft/2019-09/schema",
        "$id": "http://example.com/son",
    }
    daughter_path, daughter = tmp_path / "child/daughter.json", {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$id": "http://example.com/random/daughter",
    }
    grandchild_path, grandchild = tmp_path / "child/more/gc.json", {
        "$schema": "https://json-schema.org/draft/2019-09/schema",
        "$id": "http://example.com/also/a/grandchild",
    }

    tmp_path.joinpath("child/more").mkdir(parents=True)
    root_path.write_text(json.dumps(root))
    son_path.write_text(json.dumps(son))
    daughter_path.write_text(json.dumps(daughter))
    grandchild_path.write_text(json.dumps(grandchild))

    expected = Registry().with_contents(
        [
            (root_path.as_uri(), root),
            (son_path.as_uri(), son),
            (daughter_path.as_uri(), daughter),
            (grandchild_path.as_uri(), grandchild),
        ],
    )

    resources = loaders.from_path(tmp_path)
    registry = EMPTY_REGISTRY.with_resources(resources)
    assert registry.crawl() == expected.crawl()


def test_schema_is_inherited_downwards(tmp_path):
    root_path, root = tmp_path / "schema.json", {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$id": "http://example.com/",
    }
    son_path, son = tmp_path / "child/son.json", {
        "$id": "http://example.com/son",
    }
    daughter_path, daughter = tmp_path / "child/daughter.json", {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$id": "http://example.com/random/daughter",
    }
    grandchild_path, grandchild = tmp_path / "child/more/gc.json", {
        "$id": "http://example.com/also/a/grandchild",
    }

    tmp_path.joinpath("child/more").mkdir(parents=True)
    root_path.write_text(json.dumps(root))
    son_path.write_text(json.dumps(son))
    daughter_path.write_text(json.dumps(daughter))
    grandchild_path.write_text(json.dumps(grandchild))

    expected = Registry().with_resources(
        (each, DRAFT202012.create_resource(contents))
        for each, contents in [
            (root_path.as_uri(), root),
            (son_path.as_uri(), son),
            (daughter_path.as_uri(), daughter),
            (grandchild_path.as_uri(), grandchild),
        ]
    )

    resources = loaders.from_path(tmp_path)
    registry = EMPTY_REGISTRY.with_resources(resources)
    assert registry.crawl() == expected.crawl()


def test_empty(tmp_path):
    registry = EMPTY_REGISTRY.with_resources(loaders.from_path(tmp_path))
    assert registry == EMPTY_REGISTRY


def test_traversable(tmp_path):
    package = tmp_path / "foo"

    schemas = package / "schemas"
    path, schema = schemas / "schema.json", {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$id": "http://example.com/",
    }
    subpath, subschema = schemas / "more/child.json", {"foo": "bar"}

    schemas.joinpath("more").mkdir(parents=True)
    package.joinpath("__init__.py").touch()
    path.write_text(json.dumps(schema))
    subpath.write_text(json.dumps(subschema))

    # ?!?! -- without this, importlib.resources.files fails on 3.9 and no other
    # version!?!?
    schemas.joinpath("__init__.py").touch()

    try:
        sys.path.append(str(tmp_path))
        resources = loaders.from_traversable(files("foo.schemas"))
        registry = EMPTY_REGISTRY.with_resources(resources)

        expected = Registry().with_resources(
            (each, DRAFT202012.create_resource(contents))
            for each, contents in [
                (path.as_uri(), schema),
                (subpath.as_uri(), subschema),
            ]
        )

        assert registry.crawl() == expected.crawl()
    finally:
        sys.path.pop()



r"""
Helpers for loading `referencing.Resource`\ s out of various places.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable
import json
import os

from referencing import Resource, Specification

if TYPE_CHECKING:
    from importlib.resources.abc import Traversable

    from referencing.typing import URI


def from_path(root: Path) -> Iterable[tuple[URI, Resource[Any]]]:
    """
    Load some resources recursively from a given directory path.

    Subdirectories are defaulted to the first version seen (starting from
    the root) -- though it still is often a good idea to explicitly indicate
    what specification every resource is written for internally.
    """
    return _from_walked(_walk(root))


def _walk(path: Path) -> Iterable[Path]:
    walk = getattr(path, "walk", None)
    if walk is None:
        for dir, _, files in os.walk(path):  # pragma: no cover
            for file in files:
                yield Path(dir) / file
    else:
        for dir, _, files in walk():
            for file in files:
                yield dir / file


def _walk_traversable(root: Traversable) -> Iterable[Traversable]:
    """
    .walk() for importlib resources paths, which don't have the method :/
    """  # noqa: D415
    walking = [root]
    while walking:
        path = walking.pop()
        for each in path.iterdir():
            if each.is_dir():
                walking.append(each)
            else:
                yield each


def from_traversable(root: Traversable) -> Iterable[tuple[URI, Resource[Any]]]:
    """
    Load some resources from a given `importlib.resources` traversable.

    (I.e. load schemas from data within a Python package.)
    """
    return _from_walked(
        each
        for each in _walk_traversable(root)
        if not each.name.endswith((".py", ".pyc", ".pyo"))
    )


def _from_walked(
    paths: Iterable[Path | Traversable],
) -> Iterable[tuple[URI, Resource[Any]]]:
    specification: Specification[Any] | None = None
    for path in paths:
        contents = json.loads(path.read_text())
        if specification is None:
            specification = Specification.detect(contents)  # type: ignore[reportUnknownMemberType]
        resource = specification.detect(contents).create_resource(contents)
        uri = getattr(path, "as_uri", lambda: "")()  # pragma: no cover
        yield uri, resource
